﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using XNAContentCompiler;

namespace BatchContentCompiler
{
    class Program
    {
        static bool LoadConfiguration(string[] args)
        {
            if (args.Length != 2)
            {
                Console.WriteLine("Usage: BatchContentCompiler.exe <input path> <output path>");
                return false;
            }

            ProgramConfiguration.InputPath = args[0];
            ProgramConfiguration.OutputPath = args[1];

            return true;
        }

        static string[] _supportedExtensions = new []
            {
                "png",
                "bmp",
                "jpg",
                "tga",
                "dds",

                "wav",
                "mp3",
                "wma",

                "fx",

                "spritefont",
            };

        static string GetRelativePathFromAbsolutePath(string containerPath, string absolutePath)
        {
            if (!containerPath.EndsWith("\\"))
                containerPath += "\\";

            Uri fromUri = new Uri(containerPath);
            Uri toUri = new Uri(absolutePath);

            Uri relativeUri = fromUri.MakeRelativeUri(toUri);
            String relativePath = Uri.UnescapeDataString(relativeUri.ToString());

            return relativePath.Replace('/', Path.DirectorySeparatorChar);
        }


        static string ChecksumsPath
        {
            get
            {
                var path = Path.Combine(ProgramConfiguration.InputPath, ".checksums");

                if (!Directory.Exists(path))
                    Directory.CreateDirectory(path);

                return path;
            }
        }

        static Dictionary<string, string> GetStoredChecksums()
        {
            var checksums = new Dictionary<string, string>();

            var files = Directory.GetFiles(ChecksumsPath, "*.md5", SearchOption.AllDirectories);

            foreach (var file in files)
                 checksums.Add(Path.GetFileNameWithoutExtension(file), File.ReadAllText(file));

            return checksums;
        }

        static string ComputeFileChecksum(string fileName)
        {
            using(var s = new FileStream(fileName, FileMode.Open))
            {
                using (var md5 = MD5.Create())
                {
                    return Convert.ToBase64String(md5.ComputeHash(s));
                }
            }
        }


        static void BuildContent()
        {
            var contentBuilder = new ContentBuilder();

            contentBuilder.Clear();


            var files = Directory.GetFiles(ProgramConfiguration.InputPath, "*.*", SearchOption.AllDirectories);

            var fileNames = new HashSet<string>();

            var checksums = GetStoredChecksums();

            var processedFilesAndChecksums = new Dictionary<string, string>();

            foreach(var file in files)
            {
                if (!_supportedExtensions.Contains(Path.GetExtension(file).TrimStart('.')))
                    continue;

                var fileName = Path.GetFileName(file);

                if (fileNames.Contains(fileName))
                    throw new Exception("The file name must be unique:" + fileName);

                fileNames.Add(fileName);

                string storedChecksum = null;
                checksums.TryGetValue(fileName, out storedChecksum);

                var actualChecksum = ComputeFileChecksum(file);

                if (storedChecksum != actualChecksum ||
                    !File.Exists(Path.Combine(ProgramConfiguration.OutputPath, Path.GetFileNameWithoutExtension(fileName) + ".xnb")))
                {
                    processedFilesAndChecksums.Add(fileName, actualChecksum);
                    contentBuilder.Add(new ComboItem(fileName, file));
                }
            }

            var error = contentBuilder.Build();
            if (!string.IsNullOrEmpty(error))
            {
                Console.WriteLine(error);
                return;
            }

            string tempPath = contentBuilder.OutputDirectory;
            string[] outputFiles = Directory.GetFiles(tempPath, "*.xnb");

            if (!Directory.Exists(ProgramConfiguration.OutputPath))
                Directory.CreateDirectory(ProgramConfiguration.OutputPath);

            foreach (string file in outputFiles)
            {
                File.Copy(file, Path.Combine(ProgramConfiguration.OutputPath, Path.GetFileName(file)), true);
            }

            foreach (var p in processedFilesAndChecksums)
            {
                File.WriteAllText(Path.Combine(ChecksumsPath, p.Key + ".md5"), p.Value);
            }

            Console.WriteLine("{0} content file(s) compiled.", processedFilesAndChecksums.Count);
        }

        [STAThread]
        static int Main(string[] args)
        {
            if (!LoadConfiguration(args))
                return -1;

            try
            {
                BuildContent();

            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);

                return -1;
            }

            return 0;
        }
    }
}
